from pathlib import Path
from typing import Optional, Union
import networkx as nx

import random
import json

import numpy as np

from copy import deepcopy
from py2neo import Graph
from ipdb import set_trace

from cargo.utils import Log
from cargo.utils import ProgressBarFactory

from collections import namedtuple
from urllib.parse import urlunparse
from cargo.helper import *
from cargo.metrics import Metrics


class Cargo:
    """Context Sensitive Label Propagation for partitioning a monolith application into microservices.
    """
    verbose: bool = False

    def __init__(self, use_dgi: bool, dgi_neo4j_hostname: str, dgi_neo4j_hostport: int, dgi_neo4j_auth: str, verbose: bool):

        # namedtuple to match the internal signature of urlunparse
        Components = namedtuple(typename='Components', field_names=[
                                'scheme', 'netloc', 'url', 'path', 'query', 'fragment'])
        url = urlunparse(
            Components(
                scheme="bolt",
                netloc="%s:%s" % (dgi_neo4j_hostname, dgi_neo4j_hostport),
                query=None,
                path='',
                url='',
                fragment='')
        )
        if use_dgi:
            self.graph: nx.MultiDiGraph = self.dgi2networkx(
                url, dgi_neo4j_auth)
        Cargo.verbose = verbose

    def dgi2networkx(self, neo4j_url: str, dgi_neo4j_auth: str) -> nx.MultiDiGraph:
        """Convert the graph in DGI's neo4j to a local networkx instance.

        Args:
            neo4j_url (str): URL of the neo4j database
            dgi_neo4j_auth (str): Authentication to access the neo4j graph

        Returns:
            nx.MultiDiGraph: A networkx graph
        """
        Log.warn(
            "Loading graphs from Neo4j database. This could take a few minutes to complete.")
        graph = Graph(neo4j_url, auth=tuple(dgi_neo4j_auth.split(":")))
        # For now, we'll ignore transactional call trace
        # TODO: Add context sensitivity to transactional call trace and then use that to partition.
        cursor = graph.run(
            'Match (n:MethodNode)-'
            '[r:CALL_RETURN_DEPENDENCY|HEAP_DEPENDENCY|DATA_DEPENDENCY|TRANSACTION_READ|TRANSACTION_WRITE]-'
            '(m) Return n,r,m;')

        graph_data = cursor.data()
        Log.info("DGI graph loaded from Neo4j.")
        num_records = len(graph_data)
        Log.info("Found %d records." % num_records)

        all_contexts = []
        self.all_context_graphs = []
        self.full_G = nx.MultiDiGraph()
        self.transaction_graph = nx.MultiDiGraph()

        with ProgressBarFactory.get_progress_bar(debug=True) as p:
            for record in p.track(graph_data, total=num_records):
                if 'SQLTable' in record['n'].labels:
                    node1 = 'DATABASE_' + record['n']['name']
                    node2 = record['m']['node_method']
                    self.transaction_graph.add_edge(node2, node1)
                    continue

                if 'SQLTable' in record['m'].labels:
                    node1 = record['n']['node_method']
                    node2 = 'DATABASE_' + record['m']['name']
                    self.transaction_graph.add_edge(node1, node2)
                    continue

                node1 = record['n']['node_method']
                node2 = record['m']['node_method']

                if is_java_method(node1) or is_java_method(node2):
                    continue

                if node1 == node2:
                    continue

                def transform_ctx(ctx):
                    if isinstance(ctx, list):
                        return '[' + ', '.join(ctx) + ']'
                    else:
                        assert isinstance(ctx, str)
                        return ctx

                if 'ncontext' in record['r']:
                    ctx1 = transform_ctx(record['r']['pcontext'])
                    ctx2 = transform_ctx(record['r']['ncontext'])
                else:
                    ctx1 = transform_ctx(record['r']['context'])
                    ctx2 = transform_ctx(record['r']['context'])

                if ctx1 not in all_contexts:
                    all_contexts.append(ctx1)
                    self.all_context_graphs.append(nx.MultiDiGraph())

                if ctx2 not in all_contexts:
                    all_contexts.append(ctx2)
                    self.all_context_graphs.append(nx.MultiDiGraph())

                ctx1_num = all_contexts.index(ctx1)
                ctx2_num = all_contexts.index(ctx2)

                add_edge_weighted(self.full_G, (node1, node2, {}))

                add_edge_weighted(
                    self.all_context_graphs[ctx1_num], (node1, node2, {}))
                add_edge_weighted(
                    self.all_context_graphs[ctx2_num], (node1, node2, {}))

        for context_G in self.all_context_graphs:
            for node in self.full_G.nodes:
                if node not in context_G:
                    context_G.add_node(node)

    def assign_init_labels(self, G, init_labels, max_part, labels_file):

        partitions = nx.get_node_attributes(G, 'partition')

        if len(partitions) > 0:
            num_partitions = max(partitions.values()) + 1
        else:
            num_partitions = 0

        if init_labels == 'auto':
            running_count = 0
            if max_part is None:
                for node in G.nodes:
                    if node not in partitions:
                        G.nodes[node]['partition'] = running_count
                        running_count += 1
            else:
                assert num_partitions <= max_part

                for node in G.nodes:
                    if node not in partitions:
                        assert num_partitions <= max_part
                        G.nodes[node]['partition'] = running_count % max_part
                        running_count = (running_count + 1) % max_part

        elif init_labels == 'file':

            if labels_file is None:
                raise Exception(
                    "File name must be provided if init_labels='file'")

            assert labels_file[-5:] == '.json'

            with open(labels_file, 'r') as f:
                file_assignments = json.load(f)

            class_name_list = list(file_assignments.keys())

            for class_name in class_name_list:
                if "::" in class_name:
                    assignment = file_assignments.pop(class_name)
                    file_assignments[class_name.replace(
                        "::", "$")] = assignment

            matched = {key: False for key in file_assignments.keys()}

            for node in G.nodes:
                class_name = node.split(':')[0].strip(' <>').split('.')[-1]
                set_trace()
                if class_name in file_assignments:

                    matched[class_name] = True
                    partitions = nx.get_node_attributes(G, 'partition')

                    if node not in partitions:
                        G.nodes[node]['partition'] = file_assignments[class_name]
                    else:
                        raise Exception(
                            "If `init_labels == file`, there should be no existing partitions")

            num_matched = sum(matched.values())

            Log.info("Matched {}/{} from the partition file to {}/{} in the program graph".format(num_matched,
                                                                                                  len(file_assignments), len(nx.get_node_attributes(G, 'partition')), len(G.nodes)))

            partitions = nx.get_node_attributes(G, 'partition')
            if len(partitions.values()) != 0:
                num_part = max(partitions.values()) + 1
            else:
                num_part = 1

            if max_part is not None:
                assert num_part <= max_part

            # If too few of the nodes are labelled, random init the rest
            if (len(partitions) / float(len(G.nodes))) <= 0.10:
                raise Exception("Only {} out of {} nodes were initialized from the file".format(
                    len(partitions), len(G.nodes)))
                # Log.info("Doing random init for the remaining nodes")
                # self.assign_init_labels(G, 'auto', max_part)

        else:
            raise NotImplementedError

    def label_propagation(self, G):

        node_labels = nx.get_node_attributes(G, 'partition')

        num_partitions = max(node_labels.values()) + 1

        assert len(node_labels) == len(G.nodes)
        assert max(node_labels.values()) >= 0

        while True:
            active = False

            nodes = list(G.nodes())

            random.shuffle(nodes)

            for node in nodes:

                label_scores = {i: 0 for i in range(
                    max(node_labels.values()) + 1)}
                label_scores.update({-1: 0})

                for neighbor in G.neighbors(node):

                    edge = G.get_edge_data(node, neighbor)

                    if 'weight' in edge:
                        label_scores[node_labels[neighbor]] += edge['weight']
                    else:
                        label_scores[node_labels[neighbor]] += 1

                max_candidates = [
                    v for l, v in label_scores.items() if (l != -1) and (v != 0)]

                if len(max_candidates) != 0:
                    max_freq = max(max_candidates)
                    all_max_labels = [
                        l for l, v in label_scores.items() if v == max_freq and l != -1]

                    if node_labels[node] not in all_max_labels:
                        node_labels[node] = random.choice(all_max_labels)
                        active = True

            if not active:
                break

        # Re-number the nodes
        all_labels = set(node_labels.values())
        if -1 in all_labels:
            all_labels.remove(-1)
        label_map = {label: i for i, label in enumerate(all_labels)}
        label_map.update({-1: -1})

        for node in G.nodes:
            G.nodes[node]['partition'] = label_map[node_labels[node]]

    def prop_db(self, G):

        expanded_G = nx.MultiDiGraph(deepcopy(G))

        for edge in self.transaction_graph.edges(data=True):
            add_edge_weighted(expanded_G, (edge[0], edge[1], {}))

        expanded_G = to_undirected_simple(expanded_G)
        fill_minus_one(expanded_G)

        Log.info("Doing LPA on graph with database edges temporarily added")

        self.label_propagation(expanded_G)

        copy_partitions(expanded_G, G)

    def do_cargo(self, init_labels='auto', max_part=None, labels_file=None):
        clear_partitions(self.full_G)

        for ctx_graph in self.all_context_graphs:
            clear_partitions(ctx_graph)

        prev_graph = to_undirected_simple(self.full_G)
        self.assign_init_labels(
            prev_graph, init_labels=init_labels, max_part=max_part, labels_file=labels_file)
        fill_minus_one(prev_graph)

        if self.transaction_graph.number_of_edges() > 0:
            Log.info(
                "Found database transaction edges. Performing first round of label propogation.")
            self.prop_db(prev_graph)

        num_ctx = len(self.all_context_graphs)
        Log.info(f"Found {num_ctx} contexts.")
        ctx_order = np.random.permutation(num_ctx)

        for ctx_num in ctx_order:
            curr_graph = to_undirected_simple(self.all_context_graphs[ctx_num])

            fill_minus_one(curr_graph)
            copy_partitions(prev_graph, curr_graph)

            self.label_propagation(curr_graph)

            prev_graph = curr_graph

        labelprop_G = to_undirected_simple(self.full_G)

        fill_minus_one(labelprop_G)
        copy_partitions(curr_graph, labelprop_G)

        partition_freqs = np.array(np.unique(list(nx.get_node_attributes(
            labelprop_G, 'partition').values()), return_counts=True))
        partition_freqs = np.array(partition_freqs).T
        partition_freqs = {partition: freq for partition,
                           freq in partition_freqs if partition != -1}
        least_freq = min(partition_freqs, key=partition_freqs.get)

        # num_partitions      = max(nx.get_node_attributes(labelprop_G, 'partition').values()) + 1
        unassigned_count = 0

        for node in labelprop_G.nodes:
            if labelprop_G.nodes[node]['partition'] == -1:
                labelprop_G.nodes[node]['partition'] = least_freq
                unassigned_count += 1

        if unassigned_count > 0:
            Log.info("Warning : {} nodes out of {} were still -1 at the end of Labelprop, and so were assigned to partition {}".format(
                unassigned_count, len(labelprop_G.nodes), least_freq))

        return labelprop_G

    def compute_metrics(self, G):

        new_G = deepcopy(G)

        metrics: Metrics = Metrics(new_G, self.transaction_graph)
        static_metrics = metrics.compute_static_metrics()
        all_metrics = {**static_metrics}

        if self.transaction_graph.number_of_edges() > 0:
            all_metrics['DB'] = metrics.transaction_entropy()
        else:
            all_metrics['DB'] = 0.0

        return all_metrics

    def get_metrics_from_file(self, labels_file):

        new_G = deepcopy(self.full_G)
        clear_partitions(new_G)
        self.assign_init_labels(new_G, init_labels='file',
                                max_part=None, labels_file=labels_file)
        self.copy_new_classes_from_file(new_G, labels_file)
        metrics = self.compute_metrics(new_G)

        return metrics

    def copy_new_classes_from_file(self, labelprop_G, labels_file):

        with open(labels_file, 'r') as f:
            init_partitions = json.load(f)

        curr_partitions = nx.get_node_attributes(labelprop_G, 'partition')
        num_curr_partitions = max(curr_partitions.values()) + 1

        label_remapping = {i: i for i in range(num_curr_partitions)}
        copy_count = 0

        for class_name, partition in init_partitions.items():
            new_name = class_name.split('.')[-1].replace('::', '$')

            if not any(new_name == node.split('.')[-1] for node in labelprop_G.nodes):
                # While copying over partitions, maintain continuous numbering
                if partition not in label_remapping:
                    label_remapping[partition] = num_curr_partitions
                    num_curr_partitions += 1

                copy_count += 1
                node_name = class_name.replace('::', '$')
                labelprop_G.add_node(node_name)
                labelprop_G.nodes[node_name]['partition'] = label_remapping[partition]
                labelprop_G.nodes[node_name]['node_short_name'] = class_name

        Log.info(
            "Copied {} classes directly from the initial file".format(copy_count))

    def run(self, init_labels, max_part: Optional[int] = None, labels_file: Union[str, Path, None] = None):

        if init_labels == 'file':
            Log.info("Cargo with {} initial labels".format(labels_file))
        else:
            Log.info("Cargo with {} initial labels".format(init_labels))

        labelprop_G = self.do_cargo(init_labels, max_part, labels_file)
        assignments = nx.get_node_attributes(labelprop_G, 'partition')

        if init_labels == 'file':
            if isinstance(labels_file, str):
                labels_file = Path(labels_file)
            self.copy_new_classes_from_file(labelprop_G, labels_file)

            with open(labels_file, 'r') as f:  # type: ignore
                init_partitions = json.load(f)

            num_init_partitions = max(init_partitions.values()) + 1
            num_gen_partitions = max(assignments.values()) + 1

            Log.info("Max partitions : {}, File partitions : {}, Gen partitions : {}".format(
                max_part, num_init_partitions, num_gen_partitions))
            Log.info("Init partition sizes : {}".format(
                np.unique(list(init_partitions.values()), return_counts=True)[1]))
        else:
            num_gen_partitions = max(assignments.values()) + 1
            Log.info("Max partitions : {}, Gen partitions : {}".format(
                max_part, num_gen_partitions))

        partition_sizes = np.unique(
            list(assignments.values()), return_counts=True)[1]
        Log.info("Final partition sizes : {}".format(partition_sizes))

        metrics = self.compute_metrics(labelprop_G)

        clear_partitions(labelprop_G)

        return metrics, assignments



import logging
from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(
    level="INFO", format=FORMAT, datefmt="[%X]", handlers=[RichHandler(show_path=False)]
)


class Log(object):
    log = logging.getLogger("rich")

    @staticmethod
    def warn(msg: str):
        Log.log.warn(msg, extra={"markup": True})

    @staticmethod
    def info(msg: str):
        Log.log.info(msg, extra={"markup": True})

    @staticmethod
    def debug(msg: str):
        Log.log.debug(msg, extra={"markup": True})

    @staticmethod
    def error(msg: str):
        Log.log.error(msg, extra={"markup": True})



import re
from typing import Dict, List
from collections import defaultdict

METHOD_INFO = defaultdict()

def jsonify_context(raw: str) -> List:
        """ Convert context string to a json string.

        Args:
            ctx_list (str): Doop format context string

        Returns:
            str: JSON string representation of the doop context

        Notes:
            Take the context information from: "[class_name_1/method_name_1/obj_name_1/id1, class_name_2/method_name_2/obj_name_2/id2]"
            And, converts it to the following format:
            "{
                {
                    "class": "class_name_1",
                    "method": "method_name_1",
                    "object": "obj_name_1",
                    "instance": id1,
                },
                {
                    "class": "class_name_2",
                    "method": "method_name_2",
                    "object": "obj_name_2",
                    "instance": id2,
                }
            }"
        """

        raw_str = re.sub("[\[\]]", "", raw)
        raw_ctx_lst = raw_str.split(', ')

        for i, str_el in enumerate(raw_ctx_lst):

            ctx_dict = defaultdict(None)
            if str_el in ("<<system-thread-group>>", "<<main-thread-group>>", "<<main-thread>>", "<<immutable-context>>", "<<immutable-hcontext>>", "<<string-builder>>", "<<string-buffer>>"):
                ctx_dict["class"] = str_el
                ctx_dict["method"] = None
                ctx_dict["type"] = None
                ctx_dict["object"] = None
                ctx_dict["instance"] = 0

            elif "MockObject" in str_el:
                class_name, object_name = *str_el.split("::"),
                ctx_dict["class"] = class_name
                ctx_dict["object"] = object_name
                ctx_dict["instance"] = 0

            else:
                raw_substr = raw_ctx_lst[i].split('/')
                class_name, method_signature = *raw_substr[0][1:-1].split(": "),

                try:
                    method_rtype, method_name, _ = *re.sub("[(:)]", " ", method_signature).split(),
                except ValueError:
                    method_rtype, method_name = *re.sub("[(:)]", " ", method_signature).split(),

                object_name = raw_substr[1]
                instance_id = raw_substr[2]
                ctx_dict["class"] = class_name
                ctx_dict["method"] = method_name
                ctx_dict["type"] = method_rtype
                ctx_dict["object"] = object_name
                ctx_dict["instance"] = instance_id

            raw_ctx_lst[i] = ctx_dict

        return tuple(raw_ctx_lst)

def jsonify_method_string(raw: str) -> Dict:
        """ Convert doop style method information to a json formatted string

        Args:
            raw (str): Method information as a doop style string

        Returns:
            str: JSON string representation of the method information
        """

        # In the method info already exists, use that...
        raw = re.sub("[(:)]", " ", raw[1:-1])
        try:
            class_name, return_type, method_sig, _ = *raw.split(),
        except ValueError:
            class_name, return_type, method_sig = *raw.split(),

        key = "::".join([class_name, method_sig])
        if key in METHOD_INFO:
            method_dict = METHOD_INFO[key]
        else:
            # If not, add a new instance to the dictionary
            method_dict = {
                "name": method_sig,
                "class": class_name,
                "class_short_name": class_name.split(".")[-1],
                "return_type": return_type
            }
            METHOD_INFO[key] = method_dict

        return method_dict

def jsonify_heap_obj(heapobj_str: str) -> str:
    """ Create a JSON string from raw heap object string from doop

    Args:
        heapobj_str (str): Heap object as a string

    Returns:
        str: JSON string
    """
    raw_substr = re.sub("[<>]", "", heapobj_str)
    if raw_substr in {'string-constant', 'string-buffer', 'string-builder', 'java.lang.StringMockObject', 'null pseudo heap', 'main-thread', 'system-thread-group', 'main-thread-group'}:
        heap_obj_dict = {
            "class": raw_substr,
            "class_short_name": None,
            "method": None,
            "object": raw_substr,
            "instance": 0,
        }
    elif "MockObject" in heapobj_str:
        class_name, object_name = *heapobj_str.split("::"),
        method_name = None
        heap_obj_dict = {
            "class": class_name,
            "class_short_name": class_name.split('.')[-1],
            "method": method_name,
            "object": class_name,
            "instance": 0,
        }
    else:

        raw_substr = raw_substr.split('/')
        try:
            class_name, method_signature = raw_substr[0].split(":")
            method_signature = method_signature.strip()
        except:
            try:
                class_name, object_name = raw_substr[0].split(" ")
            except:
                if raw_substr[0].split(" ")[0] == 'method' and raw_substr[0].split(" ")[1] == 'type':
                    return {
                        "class": 'method',
                        "class_short_name": 'method',
                        "method": None,
                        "type": None,
                        "object": 'method',
                        "instance": None
                       }

            if class_name == 'class':
                return {
                    "class": object_name,
                    "class_short_name": object_name.split('.')[-1],
                    "method": None,
                    "type": None,
                    "object": object_name,
                    "instance": None
                   }

        if class_name[:7] == 'handle ':
            class_name = class_name[7:]

        try:
            method_rtype, method_name = *method_signature.split(),
        except ValueError:
            method_name = method_signature.split()[0]
            method_rtype = None,

        if len(raw_substr) == 3:
            if raw_substr[1] == 'invokedynamic_metafactory::apply':
                object_name = raw_substr[2].split(" ")[5].split(':')[0]
                instance_id = None
            else:
                object_name = raw_substr[1].split()[1]
                instance_id = raw_substr[2]
        elif len(raw_substr) == 2:
            object_name = raw_substr[1].split()[1]
            instance_id = None
        elif len(raw_substr) == 1:
            object_name = class_name
            instance_id = None
        else:
            raise Exception

        heap_obj_dict = {
            "class": class_name,
            "class_short_name": class_name.split('.')[-1],
            "method": method_name,
            "type": method_rtype,
            "object": object_name,
            "instance": instance_id
        }

    return heap_obj_dict

def gen_class_assignment(partition, bc_per_class):

    class_bcs_partition_assignment = {}
    partition_class_bcs_assignment = {}
    for key, assignment in partition.items():

        assignment = str(assignment)
        if key not in class_bcs_partition_assignment:
            class_bcs_partition_assignment[key] = {}
            class_bcs_partition_assignment[key]["business_context"] = []

        class_bcs_partition_assignment[key]['final'] = assignment
        if key in bc_per_class:
            bcs = bc_per_class[key]
        else:
            bcs = []
        class_bcs_partition_assignment[key]["business_context"].extend(bcs)

        if assignment not in partition_class_bcs_assignment:
            partition_class_bcs_assignment[assignment] = {}
            partition_class_bcs_assignment[assignment]['classes'] = []
            partition_class_bcs_assignment[assignment]['business_context'] = []

        partition_class_bcs_assignment[assignment]["classes"].append(key)
        partition_class_bcs_assignment[assignment]['business_context'].extend(bcs)
            # [bc for bc in bcs if bc not in partition_class_bcs_assignment[assignment]['business_context']])

    return class_bcs_partition_assignment, partition_class_bcs_assignment

def get_call_info(ROOT, runtime_call_volume):
    call_volume = {}
    nodes = []
    for link in runtime_call_volume:
        src = link.split("--")[0]
        tgt = link.split("--")[1]

        nodes.append(src)
        nodes.append(tgt)

        if src.lower() == str(ROOT).lower() or tgt.lower() == str(ROOT).lower():
            continue

        if src == tgt:
            continue

        #src = link['source']
        #tgt = link['target']
        call_volume[(src, tgt)] = runtime_call_volume[link]

    return list(set(nodes)), call_volume



import networkx as nx
from copy import copy, deepcopy

def is_java_method(meth_name):
    return (meth_name.split('.')[0].strip(' <>') in ['java', 'javax']) or (meth_name in ["<<string-builder>>", "<<string-constant>>", "<<null pseudo heap>>"])

def add_edge(G, full_edge):
    if full_edge not in G.edges(data=True):
        G.add_edge(full_edge[0], full_edge[1], **full_edge[2])


def add_edge_weighted(G, full_edge, interrupt=False):

    # Fix this!! (deepcopy)
    full_edge = deepcopy(full_edge)

    edge_weight = full_edge[2].pop('weight') if 'weight' in full_edge[2] else 1.0

    found_exact_edge    = False
    found_parallel_edge = False

    for u, v, data in G.edges(data=True):
        if (nx.is_directed(G) and (u, v) == (full_edge[0], full_edge[1])) or ((not nx.is_directed(G)) and (((u, v) == (full_edge[0], full_edge[1])) or ((v, u) == (full_edge[0], full_edge[1])))):

            if 'weight' in data:
                weight = data.pop('weight')
            else:
                weight = 1

            found_parallel_edge = True

            if full_edge[2] == data:
                data['weight'] = weight + edge_weight
                found_exact_edge = True
                break
            else:
                data['weight'] = weight

    if not found_exact_edge:
        if found_parallel_edge:
            if not (isinstance(G, nx.MultiDiGraph) or isinstance(G, nx.MultiGraph)):
                print("Warning : Trying to add a parallel edge, but G is not a MultiGraph")

        G.add_edge(full_edge[0], full_edge[1], **full_edge[2], weight=edge_weight)


def add_edge_no_attributes(G, full_edge):

    # Fix this!! (deepcopy)
    u, v, data = deepcopy(full_edge)
    new_weight = data.pop('weight') if 'weight' in data else 1.0
    if 'color' in data:
        data.pop('color')

    if G.has_edge(u, v):
        edge_data = G.get_edge_data(u, v)

        edge_weight = edge_data['weight'] if 'weight' in edge_data else 1.0

        G[u][v]['weight'] = edge_weight + new_weight

    else:
        G.add_edge(u, v, **data, weight=new_weight)


def to_undirected(G):
    undirected_G = nx.MultiGraph()

    for node, data in G.nodes(data=True):
        undirected_G.add_node(node, **data)

    for edge in G.edges(data=True):
        add_edge_weighted(undirected_G, edge)

    return undirected_G

def to_simple(G):
    simple_G = nx.DiGraph() if nx.is_directed(G) else nx.Graph()

    for node, data in G.nodes(data=True):
        simple_G.add_node(node, **data)

    for edge in G.edges(data=True):
        add_edge_no_attributes(simple_G, edge)

    return simple_G

def to_undirected_simple(G):
    G1 = to_undirected(G)
    G1 = to_simple(G1)
    return G1

def copy_partitions(G_from, G_to):

    for node in G_from.nodes:
        if node in G_to.nodes and 'partition' in G_from.nodes[node]:
            G_to.nodes[node]['partition'] = G_from.nodes[node]['partition']

def fill_minus_one(G):

    partitions = nx.get_node_attributes(G, 'partition')

    for node in G:
        if node not in partitions:
            G.nodes[node]['partition'] = -1

def clear_partitions(G):
    for node in G:
        if 'partition' in G.nodes[node]:
            G.nodes[node].pop('partition')

def filter_color(G, color):

    filtered_G = nx.DiGraph() if nx.is_directed(G) else nx.Graph()

    for u, v, data in G.edges(data=True):
        if data['color'] == color:
            add_edge_weighted(filtered_G, (u, v, data))

    return filtered_G

def get_unique_ordered(sequence):
    seen = set()
    return [x for x in sequence if not (x in seen or seen.add(x))]

def parse_ctx(ctx):
    contexts = ctx.split(', ')
    contexts = [c.strip().strip(' []<>') for c in contexts]

    context_list = []

    for context in contexts:
        if context in ['immutable-context', 'immutable-hcontext', 'string-builder', 'string-buffer', 'string-constant']:
            context_list.append(context)
            continue

        try:

            context_dict = {}

            if '/' in context:
                class_and_method = context.split('/')[0].split(':')

                context_dict['class']       = class_and_method[0].strip(' []<>')
                context_dict['method']      = class_and_method[1].strip(' []<>')
                context_dict['heap_obj']    = context.split('/')[1]
                context_dict['instance_id'] = context.split('/')[2]

            else:
                context_dict['class']       = context.split(':')[0]
                context_dict['heap_obj']    = context.split(':')[2]

        except:
            print("Context parsing failed")
            import pdb; pdb.set_trace()

        context_list.append(context_dict)

    assert len(context_list) == 2

    return context_list



"""
CARGO

A data-centric transformation of monoliths into microservices
"""

__version__ = "0.1.0"
__author__ = 'Rahul Krishna <imralk+oss@gmail.com> and Vikram Nitin <vikramnitin9@gmail.com>'
__credits__ = 'IBM Research :: Columbia University'

from .cargo import Cargo
__all__ = ['Cargo']



# Fancy Progress Bar
from rich.progress import (
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    TimeElapsedColumn,
    MofNCompleteColumn,
    TimeRemainingColumn
)

class ProgressBarFactory:
    @classmethod
    def get_progress_bar(cls, debug=False):
        progress_bar = Progress(
        SpinnerColumn(),
        TextColumn("•"),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        BarColumn(),
        TextColumn("• Completed/Total:"),
        MofNCompleteColumn(),
        TextColumn("• Elapsed:"),
        TimeElapsedColumn(),
        TextColumn("• Remaining:"),
        TimeRemainingColumn())
        if debug:
            progress_bar.disable = True
        return progress_bar




import re
import sys
import json
import numpy as np
import pandas as pd
import math
import networkx as nx
from pathlib import Path
from collections import defaultdict
from typing import DefaultDict, NewType, Tuple
from scipy.stats import entropy
import itertools
import os
import itertools
from copy import deepcopy

from .metric_utils import jsonify_context, jsonify_heap_obj, jsonify_method_string, gen_class_assignment, get_call_info

# -- Alisases for type hinting --
DataFrame = NewType("DataFrame", pd.DataFrame)
MultiDiGraph = NewType("MultiDiGraph", nx.classes.multidigraph.MultiDiGraph)

class Metrics:

    def __init__(self, ddg: MultiDiGraph, transaction_graph: MultiDiGraph = None):
        self.ddg = deepcopy(ddg)
        self.transaction_graph = deepcopy(transaction_graph).to_undirected()

        remapping = {}
        for node in self.ddg:
            if len(node.split('.')) > 1:
                remapping[node] = node.split('.')[-1]

        self.ddg = nx.relabel_nodes(self.ddg, remapping)
        self.transaction_graph = nx.relabel_nodes(self.transaction_graph, remapping)

        dir_path = os.path.dirname(os.path.realpath(__file__))

        self.partitions = nx.get_node_attributes(self.ddg, 'partition')

        N = max(self.partitions.values()) + 1
        assert all(i in self.partitions.values() for i in range(N))
        assert len(set(self.partitions.values())) == N

    def _static_cohesion(self) -> float:

        N = max(self.partitions.values()) + 1
        # We want partitions to be 0, 1, ..., N-1 without any missing
        assert all(i in self.partitions.values() for i in range(N))
        assert len(set(self.partitions.values())) == N

        _internal_counts = np.zeros(N, dtype=np.float32)
        _external_counts = np.zeros(N, dtype=np.float32)

        for src_class, dst_class, data in self.ddg.edges(data=True):

            if src_class not in self.partitions or dst_class not in self.partitions:
                continue

            weight = data['weight'] if 'weight' in data else 1.0

            src_partition = self.partitions[src_class]
            dst_partition = self.partitions[dst_class]

            if src_partition == dst_partition:
                _internal_counts[src_partition] += weight
            else:
                _external_counts[src_partition] += weight
                _external_counts[dst_partition] += weight

        cohesion = np.mean((2 * _internal_counts) / ((2 * _internal_counts) + _external_counts + 1e-7))

        return np.round(cohesion, 3)


    def _static_coupling(self) -> float:

        intra_partition: int = 0
        extra_partition: int = 0

        for edges in self.ddg.edges(data=True):
            src_class, dst_class, data = edges
            weight = data['weight'] if 'weight' in data else 1.0

            if (src_class not in self.partitions) or (dst_class not in self.partitions):
                if (src_class not in self.partitions) and (dst_class not in self.partitions):
                    intra_partition += weight
                else:
                    extra_partition += weight
                continue

            if self.partitions[src_class] == self.partitions[dst_class]:
                intra_partition += weight
            else:
                extra_partition += weight

        total   = extra_partition + intra_partition + 1e-7
        sipv    = extra_partition / float(total)

        return sipv

    def _business_context_purity(self, partition_class_bcs_assignment, dataset, result=None):
        #lower is better

        with open(Path(dir_path).parent.joinpath("resources/{}/bcs_per_class.json".format(dataset)), 'r') as f:
            self.bcs_per_class = json.load(f)

        self.bcs_per_class          = {key.replace("::", "$") : value for key, value in self.bcs_per_class.items()}

        """ The entropy of business context. """
        if result == None:
            result = partition_class_bcs_assignment

        e = []

        for cls, value in result.items():
            if cls == '-1':
                continue
            counts  = np.unique(value['business_context'], return_counts=True)[1]
            # freqs   = counts / np.sum(counts)
            # ent1    = np.sum(-1.0 * freqs * np.log(freqs))
            ent2    = entropy(counts)
            # assert math.abs(ent1 - ent2) < 0.001
            e.append(ent2)
        return round(np.mean(e), 3)


    def _inter_call_percentage(self, ROOT, class_bcs_partition_assignment, runtime_call_volume, dataset, result=None):
        """ The percentage of runtime call between two clusters. """
        #lower is better
        with open(Path(dir_path).parent.joinpath("resources/{}/runtime_call_volume.json".format(dataset)), 'r') as f:
            self.runtime_call_volume = json.load(f)
        self.runtime_call_volume    = {key.replace("::", "$") : value for key, value in self.runtime_call_volume.items()}

        if result == None:
            result = class_bcs_partition_assignment

        n_total = 0
        n_inter = 0
        for call, volume in runtime_call_volume.items():
            src, target = call.split("--")
            if src.lower() == str(ROOT).lower() or target.lower() == str(ROOT).lower():
                continue

            if src == target:
                continue

            if src and target:
                try:
                    src_assignment, target_assignment = result[src]['final'], result[target]['final']
                    n_total += volume
                    if src_assignment != target_assignment:
                        n_inter += volume
                except KeyError:
                    continue

        try:
            r = n_inter * 1.0 / n_total
        except ZeroDivisionError:
            r = float("Inf")

        return round(r, 3)


    def transaction_entropy(self):

        partition_lists = []
        _entropy = []

        if self.transaction_graph.is_directed():
            self.transaction_graph = self.transaction_graph.to_undirected()

        for node in self.transaction_graph.nodes:
            if 'DATABASE_' in node:
                partition_list = []
                for neighbor in self.transaction_graph.neighbors(node):
                    if neighbor in self.partitions:
                        partition_list += [self.partitions[neighbor]]
                    else:
                        partition_list += [-1]

                partition_lists += [partition_list]
                counts = np.unique(partition_list, return_counts=True)[1]
                _entropy += [entropy(counts)]

        return np.round(np.mean(_entropy), 3)


    def compute_dataflow_metrics(self, dataset) -> dict:

        class_bcs_partition_assignment, partition_class_bcs_assignment = gen_class_assignment(self.partitions,self.bcs_per_class)

        bcp = self._business_context_purity(partition_class_bcs_assignment, dataset)
        icp = self._inter_call_percentage('Root', class_bcs_partition_assignment, self.runtime_call_volume, dataset)

        dataflow_metrics    = {'BCP'    : bcp,
                               'ICP'    : icp}

        return dataflow_metrics

    def compute_static_metrics(self) -> dict:
        """ Compute all static metrics

        Returns:
            DataFrame: A summary of the metrics
        """

        static_metrics = {'Cohesion'    : self._static_cohesion(),
                          'Coupling'    : self._static_coupling()}

        return static_metrics



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='minerva-cargo',
    version='1.0',
    url="https://github.com/ibm/codenet-minerva-cargo",
    author="Rahul Krishna",
    description="A data-centric transformation of monoliths into microservices",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    install_requires=[
        'scipy',
        'py2neo',
        'numpy',
        'pandas',
        'networkx',
        'tqdm',
        'scikit-learn',
        'tackle-dgi',
        'rich'
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
    extras_require={
        "dev": [
            "nose==1.3.7",
            "pinocchio==0.4.3",
            "coverage==6.3.2",
            "pylint==2.13",
            "ipdb"
        ],
    }
)
